@@ -1,14 +1,14 @@
 /*
  *  hdbc.c
  *
- *  $Id: hdbc.c,v 1.30 2006/01/20 15:58:34 source Exp $
+ *  $Id: hdbc.c,v 1.41 2009/08/30 10:48:33 source Exp $
  *
  *  Data source connect object management functions
  *
  *  The iODBC driver manager.
  *
  *  Copyright (C) 1995 by Ke Jin <kejin@empress.com>
- *  Copyright (C) 1996-2006 by OpenLink Software <iodbc@openlinksw.com>
+ *  Copyright (C) 1996-2009 by OpenLink Software <iodbc@openlinksw.com>
  *  All Rights Reserved.
  *
  *  This software is released under the terms of either of the following
@@ -96,7 +96,7 @@
 #include <stdio.h>
 
 
-extern SQLRETURN _iodbcdm_driverunload (HDBC hdbc);
+extern SQLRETURN _iodbcdm_driverunload (HDBC hdbc, int ver);
 
 static SQLRETURN
 _iodbcdm_drvopt_store (SQLHDBC hdbc, SQLUSMALLINT fOption, SQLULEN vParam,
@@ -193,6 +193,17 @@ SQLAllocConnect_Internal (
   if (genv->odbc_ver == 0)
     genv->odbc_ver = SQL_OV_ODBC2;
   pdbc->hdesc = NULL;
+
+  pdbc->cp_pdbc = NULL;
+  pdbc->cp_in_use = FALSE;
+  pdbc->cp_timeout = 0;
+  pdbc->cp_expiry_time = 0;
+  pdbc->cp_retry_wait = 0;
+  pdbc->cp_probe = NULL;
+  pdbc->cp_dsn = NULL;
+  pdbc->cp_uid = NULL;
+  pdbc->cp_pwd = NULL;
+  pdbc->cp_connstr = NULL;
 #endif
   pdbc->genv = genv;
 
@@ -250,7 +261,7 @@ SQLAllocConnect (SQLHENV henv, SQLHDBC * phdbc)
 
 
 SQLRETURN
-SQLFreeConnect_Internal (SQLHDBC hdbc)
+SQLFreeConnect_Internal (SQLHDBC hdbc, int ver)
 {
   CONN (pdbc, hdbc);
   GENV (genv, pdbc->genv);
@@ -279,7 +290,7 @@ SQLFreeConnect_Internal (SQLHDBC hdbc)
     }
 
   /* free this dbc */
-  _iodbcdm_driverunload (pdbc);
+  _iodbcdm_driverunload (pdbc, 3);
 
   /* free driver connect options */
   _iodbcdm_drvopt_free (pdbc);
@@ -299,7 +310,7 @@ SQLFreeConnect (SQLHDBC hdbc)
   ENTER_HDBC (hdbc, 1,
     trace_SQLFreeConnect (TRACE_ENTER, hdbc));
 
-  retcode = SQLFreeConnect_Internal (hdbc);
+  retcode = SQLFreeConnect_Internal (hdbc, 2);
 
   LEAVE_HDBC (hdbc, 1,
     trace_SQLFreeConnect (TRACE_LEAVE, hdbc);
@@ -318,9 +329,15 @@ _iodbcdm_SetConnectOption (
   CONN (pdbc, hdbc);
   ENVR (penv, pdbc->henv);
   STMT (pstmt, NULL);
-  HPROC hproc = SQL_NULL_HPROC;
+  HPROC hproc2 = SQL_NULL_HPROC;
+  HPROC hproc3 = SQL_NULL_HPROC;
   sqlstcode_t sqlstat = en_00000;
   SQLRETURN retcode = SQL_SUCCESS;
+  SQLUINTEGER odbc_ver;
+  SQLUINTEGER dodbc_ver;
+
+  odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
+  dodbc_ver = (penv != SQL_NULL_HENV) ? penv->dodbc_ver : odbc_ver;
 
 #if (ODBCVER < 0x0300)
   /* check option */
@@ -474,7 +491,6 @@ _iodbcdm_SetConnectOption (
        * been established.  
        */
      void * _vParam = NULL;
-     int procid = 0;
 
      if ((penv->unicode_driver && waMode != 'W') 
         || (!penv->unicode_driver && waMode == 'W'))
@@ -504,28 +520,34 @@ _iodbcdm_SetConnectOption (
           }
       }
 
-#if (ODBCVER >= 0x0300)
      if (penv->unicode_driver)
        {
          /* SQL_XXX_W */
-         if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectAttrW))
-             != SQL_NULL_HPROC)
-           procid = en_SetConnectAttrW;
+#if (ODBCVER >= 0x0300)
+         hproc3 = _iodbcdm_getproc (pdbc, en_SetConnectAttrW);
+#endif
+         hproc2 = _iodbcdm_getproc (pdbc, en_SetConnectOptionW);
        }
      else
        {
          /* SQL_XXX */
          /* SQL_XXX_A */
-         if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectAttr))
-             != SQL_NULL_HPROC)
-           procid = en_SetConnectAttr;
-         else
-         if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectAttrA))
-             != SQL_NULL_HPROC)
-           procid = en_SetConnectAttrA;
+#if (ODBCVER >= 0x0300)
+         hproc3 = _iodbcdm_getproc (pdbc, en_SetConnectAttr);
+	 if (hproc3 == SQL_NULL_HPROC)
+	   hproc3 = _iodbcdm_getproc (pdbc, en_SetConnectAttrA);
+#endif
+         hproc2 = _iodbcdm_getproc (pdbc, en_SetConnectOption);
+	 if (hproc2 == SQL_NULL_HPROC)
+	   hproc2 = _iodbcdm_getproc (pdbc, en_SetConnectOptionA);
        }
 
-      if (hproc != SQL_NULL_HPROC)
+      if (odbc_ver == SQL_OV_ODBC2 && (dodbc_ver == SQL_OV_ODBC2
+	      || (dodbc_ver == SQL_OV_ODBC3 && hproc2 != SQL_NULL_HPROC)))
+	hproc3 = SQL_NULL_HPROC;
+
+#if (ODBCVER >= 0x0300)
+      if (hproc3 != SQL_NULL_HPROC)
 	{
 	  switch (fOption)
 	    {
@@ -538,7 +560,7 @@ _iodbcdm_SetConnectOption (
 	    case SQL_ATTR_QUIET_MODE:
 	    case SQL_ATTR_TRANSLATE_OPTION:
 	    case SQL_ATTR_TXN_ISOLATION:
-	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+	      CALL_DRIVER (hdbc, pdbc, retcode, hproc3,
 		  (pdbc->dhdbc, fOption, vParam, 0));
 	      break;
 
@@ -554,46 +576,26 @@ _iodbcdm_SetConnectOption (
 
 	    default:		/* string & driver defined */
 
-	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+	      CALL_DRIVER (hdbc, pdbc, retcode, hproc3,
 		  (pdbc->dhdbc, fOption, vParam, SQL_NTS));
 	    }
 	}
       else
 #endif
 	{
-          if (penv->unicode_driver)
-            {
-             /* SQL_XXX_W */
-             if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectOptionW))
-                 != SQL_NULL_HPROC)
-               procid = en_SetConnectOptionW;
-            }
-          else
-            {
-             /* SQL_XXX */
-             /* SQL_XXX_A */
-             if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectOption))
-                != SQL_NULL_HPROC)
-               procid = en_SetConnectOption;
-             else
-             if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectOptionA))
-                != SQL_NULL_HPROC)
-               procid = en_SetConnectOptionA;
-            }
-
-	  if (hproc == SQL_NULL_HPROC)
+	  if (hproc2 == SQL_NULL_HPROC)
 	    {
 	      PUSHSQLERR (pdbc->herr, en_IM001);
 	      MEM_FREE(_vParam);
 	      return SQL_ERROR;
 	    }
 
-	  CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+	  CALL_DRIVER (hdbc, pdbc, retcode, hproc2,
 	      (pdbc->dhdbc, fOption, vParam));
 	}
 
       MEM_FREE(_vParam);
-      if (retcode != SQL_SUCCESS && retcode != SQL_SUCCESS_WITH_INFO)
+      if (!SQL_SUCCEEDED (retcode))
 	{
 	  return retcode;
 	}
@@ -702,12 +704,12 @@ SQLSetConnectOption (
   SQLUSMALLINT		  fOption,
   SQLULEN		  vParam)
 {
-  ENTER_HDBC (hdbc, 1,
+  ENTER_HDBC (hdbc, 0,
     trace_SQLSetConnectOption (TRACE_ENTER, hdbc, fOption, vParam));
 
   retcode = _iodbcdm_SetConnectOption (hdbc, fOption, vParam, 'A');
 
-  LEAVE_HDBC (hdbc, 1,
+  LEAVE_HDBC (hdbc, 0,
     trace_SQLSetConnectOption (TRACE_LEAVE, hdbc, fOption, vParam));
 }
 
@@ -719,12 +721,12 @@ SQLSetConnectOptionA (
   SQLUSMALLINT		  fOption,
   SQLULEN		  vParam)
 {
-  ENTER_HDBC (hdbc, 1,
+  ENTER_HDBC (hdbc, 0,
     trace_SQLSetConnectOption (TRACE_ENTER, hdbc, fOption, vParam));
 
   retcode = _iodbcdm_SetConnectOption (hdbc, fOption, vParam, 'A');
 
-  LEAVE_HDBC (hdbc, 1,
+  LEAVE_HDBC (hdbc, 0,
     trace_SQLSetConnectOption (TRACE_LEAVE, hdbc, fOption, vParam));
 }
 
@@ -735,12 +737,12 @@ SQLSetConnectOptionW (
   SQLUSMALLINT		  fOption,
   SQLULEN		  vParam)
 {
-  ENTER_HDBC (hdbc, 1,
+  ENTER_HDBC (hdbc, 0,
     trace_SQLSetConnectOptionW (TRACE_ENTER, hdbc, fOption, vParam));
 
   retcode = _iodbcdm_SetConnectOption (hdbc, fOption, vParam, 'W');
 
-  LEAVE_HDBC (hdbc, 1,
+  LEAVE_HDBC (hdbc, 0,
     trace_SQLSetConnectOptionW (TRACE_LEAVE, hdbc, fOption, vParam));
 }
 #endif
@@ -755,9 +757,15 @@ _iodbcdm_GetConnectOption (
 {
   CONN (pdbc, hdbc);
   ENVR (penv, pdbc->henv);
-  HPROC hproc = SQL_NULL_HPROC;
+  HPROC hproc2 = SQL_NULL_HPROC;
+  HPROC hproc3 = SQL_NULL_HPROC;
   sqlstcode_t sqlstat = en_00000;
   SQLRETURN retcode = SQL_SUCCESS;
+  SQLUINTEGER odbc_ver;
+  SQLUINTEGER dodbc_ver;
+
+  odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
+  dodbc_ver = (penv != SQL_NULL_HENV) ? penv->dodbc_ver : odbc_ver;
 
 #if (ODBCVER < 0x0300)
   /* check option */
@@ -853,7 +861,6 @@ _iodbcdm_GetConnectOption (
 
       void * _Param = NULL;
       void * paramOut = pvParam;
-      int procid = 0;
 
       switch (fOption)
         {
@@ -886,29 +893,35 @@ _iodbcdm_GetConnectOption (
           break;
         }
 
-#if (ODBCVER >= 0x0300)
 
      if (penv->unicode_driver)
        {
          /* SQL_XXX_W */
-         if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectAttrW))
-             != SQL_NULL_HPROC)
-           procid = en_GetConnectAttrW;
+#if (ODBCVER >= 0x0300)
+         hproc3 = _iodbcdm_getproc (pdbc, en_GetConnectAttrW);
+#endif
+         hproc2 = _iodbcdm_getproc (pdbc, en_GetConnectOptionW);
        }
      else
        {
          /* SQL_XXX */
          /* SQL_XXX_A */
-         if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectAttr))
-             != SQL_NULL_HPROC)
-           procid = en_GetConnectAttr;
-         else
-         if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectAttrA))
-             != SQL_NULL_HPROC)
-           procid = en_GetConnectAttrA;
+#if (ODBCVER >= 0x0300)
+         hproc3 = _iodbcdm_getproc (pdbc, en_GetConnectAttr);
+	 if (hproc3 == SQL_NULL_HPROC)
+           hproc3 = _iodbcdm_getproc (pdbc, en_GetConnectAttrA);
+#endif
+         hproc2 = _iodbcdm_getproc (pdbc, en_GetConnectOption);
+	 if (hproc2 == SQL_NULL_HPROC)
+           hproc2 = _iodbcdm_getproc (pdbc, en_GetConnectOptionA);
        }
 
-      if (hproc != SQL_NULL_HPROC)
+      if (odbc_ver == SQL_OV_ODBC2 && (dodbc_ver == SQL_OV_ODBC2
+	      || (dodbc_ver == SQL_OV_ODBC3 && hproc2 != SQL_NULL_HPROC)))
+	hproc3 = SQL_NULL_HPROC;
+
+#if (ODBCVER >= 0x0300)
+      if (hproc3 != SQL_NULL_HPROC)
 	{
 	  switch (fOption)
 	    {
@@ -921,7 +934,7 @@ _iodbcdm_GetConnectOption (
 	    case SQL_ATTR_QUIET_MODE:
 	    case SQL_ATTR_TRANSLATE_OPTION:
 	    case SQL_ATTR_TXN_ISOLATION:
-	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+	      CALL_DRIVER (hdbc, pdbc, retcode, hproc3,
 		  (pdbc->dhdbc, fOption, paramOut, 0, NULL));
 	      break;
 
@@ -937,7 +950,7 @@ _iodbcdm_GetConnectOption (
 
 	    default:		/* string & driver defined */
 
-	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+	      CALL_DRIVER (hdbc, pdbc, retcode, hproc3,
 		  (pdbc->dhdbc, fOption, paramOut, SQL_MAX_OPTION_STRING_LENGTH, NULL));
 
 	    }
@@ -945,40 +958,21 @@ _iodbcdm_GetConnectOption (
       else
 #endif
 	{
-          if (penv->unicode_driver)
-            {
-             /* SQL_XXX_W */
-             if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectOptionW))
-                 != SQL_NULL_HPROC)
-               procid = en_GetConnectOptionW;
-            }
-          else
-            {
-             /* SQL_XXX */
-             /* SQL_XXX_A */
-             if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectOption))
-                != SQL_NULL_HPROC)
-               procid = en_GetConnectOption;
-             else
-             if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectOptionA))
-                != SQL_NULL_HPROC)
-               procid = en_GetConnectOptionA;
-            }
-
-	  if (hproc == SQL_NULL_HPROC)
+	  if (hproc2 == SQL_NULL_HPROC)
 	    {
 	      PUSHSQLERR (pdbc->herr, en_IM001);
               MEM_FREE(_Param);
 	      return SQL_ERROR;
 	    }
 
-	  CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+	  CALL_DRIVER (hdbc, pdbc, retcode, hproc2,
 	      (pdbc->dhdbc, fOption, paramOut));
 	}
 
+
       if (pvParam
-          && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
-          &&  ((penv->unicode_driver && waMode != 'W') 
+          && SQL_SUCCEEDED (retcode)
+          && ((penv->unicode_driver && waMode != 'W') 
               || (!penv->unicode_driver && waMode == 'W')))
         {
           switch (fOption)
@@ -1053,12 +1047,12 @@ SQLGetConnectOptionA (
   SQLUSMALLINT		  fOption,
   SQLPOINTER		  pvParam)
 {
-  ENTER_HDBC (hdbc, 1,
+  ENTER_HDBC (hdbc, 0,
     trace_SQLGetConnectOption (TRACE_ENTER, hdbc, fOption, pvParam));
 
   retcode = _iodbcdm_GetConnectOption (hdbc, fOption, pvParam, 'A');
 
-  LEAVE_HDBC (hdbc, 1,
+  LEAVE_HDBC (hdbc, 0,
     trace_SQLGetConnectOption (TRACE_LEAVE, hdbc, fOption, pvParam));
 }
 
@@ -1069,12 +1063,12 @@ SQLGetConnectOptionW (
   SQLUSMALLINT		  fOption,
   SQLPOINTER		  pvParam)
 {
-  ENTER_HDBC (hdbc, 1,
+  ENTER_HDBC (hdbc, 0,
     trace_SQLGetConnectOptionW (TRACE_ENTER, hdbc, fOption, pvParam));
 
   retcode = _iodbcdm_GetConnectOption (hdbc, fOption, pvParam, 'W');
 
-  LEAVE_HDBC (hdbc, 1,
+  LEAVE_HDBC (hdbc, 0,
     trace_SQLGetConnectOptionW (TRACE_LEAVE, hdbc, fOption, pvParam));
 }
 #endif
@@ -1086,9 +1080,16 @@ _iodbcdm_transact (
     UWORD fType)
 {
   CONN (pdbc, hdbc);
+  ENVR (penv, pdbc->henv);
   STMT (pstmt, NULL);
-  HPROC hproc;
+  HPROC hproc2 = SQL_NULL_HPROC;
+  HPROC hproc3 = SQL_NULL_HPROC;
   SQLRETURN retcode;
+  SQLUINTEGER odbc_ver;
+  SQLUINTEGER dodbc_ver;
+
+  odbc_ver = ((GENV_t *) pdbc->genv)->odbc_ver;
+  dodbc_ver = (penv != SQL_NULL_HENV) ? penv->dodbc_ver : odbc_ver;
 
   /* check state */
   switch (pdbc->state)
@@ -1106,56 +1107,56 @@ _iodbcdm_transact (
       break;
     }
 
-  for (pstmt = (STMT_t *) (pdbc->hstmt);
-      pstmt != NULL;
-      pstmt = pstmt->next)
+  for (pstmt = (STMT_t *) (pdbc->hstmt); pstmt != NULL; pstmt = pstmt->next)
     {
-      if (pstmt->state >= en_stmt_needdata
-	  || pstmt->asyn_on != en_NullProc)
+      if (pstmt->state >= en_stmt_needdata || pstmt->asyn_on != en_NullProc)
 	{
 	  PUSHSQLERR (pdbc->herr, en_S1010);
 
 	  return SQL_ERROR;
 	}
     }
-  hproc = _iodbcdm_getproc (pdbc, en_Transact);
 
-  if (hproc != SQL_NULL_HPROC)
+
+  hproc2 = _iodbcdm_getproc (pdbc, en_Transact);
+#if (ODBCVER >= 0x300)
+  hproc3 = _iodbcdm_getproc (pdbc, en_EndTran);
+#endif
+
+  if (odbc_ver == SQL_OV_ODBC2 &&
+      (dodbc_ver == SQL_OV_ODBC2
+	  || (dodbc_ver == SQL_OV_ODBC3 && hproc2 != SQL_NULL_HPROC)))
+    hproc3 = SQL_NULL_HPROC;
+
+
+#if (ODBCVER >= 0x300)
+  if (hproc3 != SQL_NULL_HPROC)
     {
-      CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_Transact,
+      CALL_DRIVER (hdbc, pdbc, retcode, hproc3,
+	  (SQL_HANDLE_DBC, pdbc->dhdbc, fType));
+    }
+  else 
+#endif
+    {
+      if (hproc2 == SQL_NULL_HPROC)
+        {
+          PUSHSQLERR (pdbc->herr, en_IM001);
+          return SQL_ERROR;
+        }
+      CALL_DRIVER (hdbc, pdbc, retcode, hproc2,
 	  (SQL_NULL_HENV, pdbc->dhdbc, fType));
     }
-  else
-    {
-#if (ODBCVER >= 0x300)
-      hproc = _iodbcdm_getproc (pdbc, en_EndTran);
-      if (hproc != SQL_NULL_HPROC)
-	{
-	  CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_EndTran,
-	      (SQL_HANDLE_DBC, pdbc->dhdbc, fType));
-	}
-      else
-	{
-	  PUSHSQLERR (pdbc->herr, en_IM001);
-	  return SQL_ERROR;
-	}
-#else
-      PUSHSQLERR (pdbc->herr, en_IM001);
-      return SQL_ERROR;
-#endif
-    }
+
 
   /* state transition */
-  if (retcode != SQL_SUCCESS && retcode != SQL_SUCCESS_WITH_INFO)
+  if (!SQL_SUCCEEDED (retcode))
     {
       return retcode;
     }
 
   pdbc->state = en_dbc_hstmt;
 
-  for (pstmt = (STMT_t *) (pdbc->hstmt);
-      pstmt != NULL;
-      pstmt = pstmt->next)
+  for (pstmt = (STMT_t *) (pdbc->hstmt); pstmt != NULL; pstmt = pstmt->next)
     {
       switch (pstmt->state)
 	{
@@ -1198,10 +1199,8 @@ _iodbcdm_transact (
 	      if (pdbc->cb_commit == SQL_CB_CLOSE
 		  || pdbc->cb_rollback == SQL_CB_CLOSE)
 		{
-		  pstmt->state
-		      = en_stmt_prepared;
-		  pstmt->cursor_state
-		      = en_stmt_cursor_no;
+		  pstmt->state = en_stmt_prepared;
+		  pstmt->cursor_state = en_stmt_cursor_no;
 		  break;
 		}
 	      break;
@@ -1263,7 +1262,7 @@ SQLTransact_Internal (
 	}
     }
 
-  if (retcode != SQL_SUCCESS && retcode != SQL_SUCCESS_WITH_INFO)
+  if (!SQL_SUCCEEDED (retcode))
     {
       /* fail on one of the connection */
       return SQL_ERROR;
